//
// (C) 2009 Mike Brent aka Tursi aka HarmlessLion.com
// This software is provided AS-IS. No warranty
// express or implied is provided.
//
// This notice defines the entire license for this code.
// All rights not explicity granted here are reserved by the
// author.
//
// You may redistribute this software provided the original
// archive is UNCHANGED and a link back to my web page,
// http://harmlesslion.com, is provided as the author's site.
// It is acceptable to link directly to a subpage at harmlesslion.com
// provided that page offers a URL for that purpose
//
// Source code, if available, is provided for educational purposes
// only. You are welcome to read it, learn from it, mock
// it, and hack it up - for your own use only.
//
// Please contact me before distributing derived works or
// ports so that we may work out terms. I don't mind people
// using my code but it's been outright stolen before. In all
// cases the code must maintain credit to the original author(s).
//
// -COMMERCIAL USE- Contact me first. I didn't make
// any money off it - why should you? ;) If you just learned
// something from this, then go ahead. If you just pinched
// a routine or two, let me know, I'll probably just ask
// for credit. If you want to derive a commercial tool
// or use large portions, we need to talk. ;)
//
// If this, itself, is a derived work from someone else's code,
// then their original copyrights and licenses are left intact
// and in full force.
//
// http://harmlesslion.com - visit the web page for contact info
//
//*****************************************************
//* Ami 99 - TI Emulator for Win32					  *
//* by M.Brent                                        *
//* Win32 WindowProc                                  *
//*****************************************************

#define WIN32_LEAN_AND_MEAN
#define _WIN32_WINNT 0x0400

#include <stdio.h>
#include <windows.h>
#include <malloc.h>
#include <ddraw.h>
#include <commctrl.h>
#include <commdlg.h>
#include <atlstr.h>

#include "resource.h"
#include "tiemul.h"
#include "keyboard/kb.h"

extern const char *szDefaultWindowText;
extern HDC tmpDC;
extern int VDPDebug;
extern int max_volume;
extern int starttimer9901;
extern int timer9901;										// 9901 interrupt timer
extern int timer9901IntReq;
extern int timer9901IntOk;
extern int nSystem;
extern int nCartGroup;
extern int nCart;
extern int keyboard, ps2keyboardok;
extern bool fKeyEverPressed;
extern void GenerateToneBuffer();
extern void GenerateSIDBuffer();
extern int max_cpf, oldmax;
// needed for configuration
extern char AVIFileName[256];
extern int drawspeed, fJoy, joy1mode, joy2mode;
extern int fJoystickActiveOnKeys;
extern int slowdown_keyboard;
extern HINSTANCE hInstance;						// global program instance
extern int TVFiltersAvailable;
extern int TVScanLines;
extern volatile signed long cycles_left;		// runs the CPU throttle
extern struct CARTS Users[];
extern struct _break BreakPoints[];
extern int nBreakPoints;
extern char lines[34][DEBUGLEN];				// debug lines
extern bool bDebugDirty;
extern Word disasm[20];							// last 20 addresses for disasm
extern bool bScrambleMemory;
// VDP tables
extern int SIT, CT, PDT, SAL, SDT, CTsize, PDTsize;
extern bool CPUSpeechHalt;
extern Byte CPUSpeechHaltByte;
extern int cpucount, cpuframes;					// CPU counters for timing
extern int timercount;							// Used to estimate runtime
extern bool bDisableBlank, bDisableSprite, bDisableBackground;
// sams config
extern int sams_enabled, sams_size;
extern Byte staticCPU[0x10000];					// main memory for debugger
// sound
extern int nRegister[4];						// frequency registers
extern int nVolume[4];							// volume attenuation
// back buffer for sizing
extern DDSURFACEDESC2 CurrentDDSD;
// debugger
extern HWND hBugWnd;

extern int nCycleCount;
extern void sound_init(int freq);
extern void SetSoundVolumes();
extern void MuteAudio();

extern void (*InitSid)();
extern void (*sid_update)(short *buf, double nAudioIn, int nSamples);
extern void (*write_sid)(Word ad, Byte dat);
extern void (*SetSidFrequency)(int);
extern void (*SetSidEnable)(bool);
extern void (*SetSidBanked)(bool);
extern bool (*GetSidEnable)(void);

const char *pCurrentHelpMsg=NULL;
HWND hKBMap=NULL;
HWND hHeatMap=NULL;
HWND hBrkHlp=NULL;
HWND hTVDlg=NULL;
HBITMAP hHeatBmp=NULL;

#ifndef GET_X_LPARAM
#define GET_X_LPARAM(x) (x&0xffff)
#endif
#ifndef GET_Y_LPARAM
#define GET_Y_LPARAM(y) (y>>16)
#endif

static LONG nOldVolume[11]={ -10000, -10000, -10000, -10000, -10000, -10000, -10000, -10000, -10000, -10000, -10000};

// Stuff for the debug window
HANDLE hDebugWindowUpdateEvent = CreateEvent(NULL, false, false, NULL);
static int nMemType=0;
static bool bFrozenText=false;
// top addresses for the memory banks
static char szTopMemory[5][32] = { "", "", "R0", "R0", "R0" };		// CPU, VDP, GROM - hex address or Register number (Rx)

// references
BOOL CALLBACK DebugBoxProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK BreakPointHelpProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

/////////////////////////////////////////////////////////////////////////
// Window handler
/////////////////////////////////////////////////////////////////////////
LONG FAR PASCAL myproc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	// winuser.h has the VK_key key defines
	// also fill in the key[] array for on/off
    
	PAINTSTRUCT ps;
    HDC hDC;
	char szTemp[1024];
	float height;
	int i1;
	RECT myrect, myrect2;

	if (myWnd == hwnd) {	// Main TI window
		switch(msg) {
		case WM_INITMENUPOPUP:
//			MuteAudio();		// stop audio during menu
			if (IsClipboardFormatAvailable(CF_TEXT)) {
				EnableMenuItem(GetMenu(myWnd), ID_EDITPASTE, MF_ENABLED | MF_BYCOMMAND);
			} else {
				EnableMenuItem(GetMenu(myWnd), ID_EDITPASTE, MF_GRAYED | MF_BYCOMMAND);
			}
			break;

		case WM_UNINITMENUPOPUP:
			// WinVer 0x0500 and above only ;)
			SetSoundVolumes();
			break;

		case WM_DISPLAYCHANGE:
			if (NULL != tmpDC) {
				DeleteDC(tmpDC);
				tmpDC=CreateCompatibleDC(NULL);
			}
			break;
		
		case WM_PAINT:
			hDC = BeginPaint(hwnd, &ps);
			GetClientRect(myWnd, &myrect);
			if (StretchMode==0) {
				FillRect(hDC, &myrect, (HBRUSH)(COLOR_MENU+1));
			}
			SetEvent(BlitEvent);
			EndPaint(hwnd, &ps);
			break;

		case WM_DESTROY:
			quitflag=1;
			PostQuitMessage(0);
			break;
		
		case WM_KEYDOWN:
			key[wParam]=1;
			if (lParam&0x1000000) {
				decode(0xe0);	// extended
			}
			decode(wParam);
			fKeyEverPressed=true;
			break;

		case WM_KEYUP:
			key[wParam]=0;
			if (lParam&0x1000000) {
				decode(0xe0);	// extended
			}
			decode(0xf0);	// key up
			decode(wParam);
			break;

		case WM_SYSKEYDOWN:				// returns from ALT and ALT+KEY (I use as FCTN)
			// some system keys we want Windows to process, namely F4 (close)
			if ((wParam != VK_F4) & (wParam != VK_RETURN))
			{	
				key[wParam]=1;
				if (lParam&0x1000000) {
					decode(0xe0);	// extended
				}
				decode(wParam);
			}
			else
			{
				switch (wParam)
				{
				case VK_F4:
					quitflag=1;
					PostQuitMessage(0);
					break;
				
				default:
					return(DefWindowProc(hwnd, msg, wParam, lParam));
				}
			}
			break;

		case WM_SYSKEYUP:
			key[wParam]=0;
			if (lParam&0x1000000) {
				decode(0xe0);	// extended
			}
			decode(0xf0);	// key up
			decode(wParam);
			break;

		case WM_SYSCHAR:
			// Don't remove this check, even if we need no ALT keys - otherwise all FCTN keys on the TI ding ;)
			// Fullscreen toggle - Alt-Enter
			if ((wParam==VK_RETURN)&&((lParam&0x8000)==0)) {
				if (3 == StretchMode) {
					StretchMode=2;
					PostMessage(hwnd, WM_COMMAND, ID_VIDEO_STRETCHMODE_NONE+StretchMode, 1);
				} else {
					if ((2 == StretchMode) && (0 != FullScreenMode)) {
						StretchMode=3;
						PostMessage(hwnd, WM_COMMAND, ID_VIDEO_STRETCHMODE_DXFULL_320X240X8+FullScreenMode-1, 1);
					}
				}
			}
			break;

		case WM_LBUTTONDBLCLK: 
			{
				// If in text mode, paste the character under the mouse cursor to the TI
				char ch;
				if (NULL != PasteString) {
					Beep(550,100);
					break;
				}
				GetClientRect(myWnd, &myrect);
				ch=VDPGetChar(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam), myrect.right-myrect.left, myrect.bottom-myrect.top);
				if (-1 != ch) {
					PasteString=(char*)malloc(2);
					PasteString[0]=ch;
					PasteString[1]='\0';
					PasteIndex=PasteString;
				}
			}
			break;

		case WM_COMMAND:
			// silence in case this takes a while
//			MuteAudio();
			// Check for dynamic ones first, so we don't need a huge switch
			if ((wParam >= ID_SYSTEM_0) && (wParam < ID_SYSTEM_0+100)) {
				// user requested to change system
				int ret;
				if ((!lParam)&&(fKeyEverPressed)) {
					ret=MessageBox(hwnd, "This will reset the emulator - are you sure?", "Change System Type", MB_YESNO|MB_ICONQUESTION);
				} else {
					ret=IDYES;
				}
				if (IDYES == ret) {
					nSystem=wParam-ID_SYSTEM_0;
					for (int idx=0; idx<100; idx++) {
						if (idx == nSystem) {
							CheckMenuItem(GetMenu(myWnd), ID_SYSTEM_0+idx, MF_CHECKED);
						} else {
							CheckMenuItem(GetMenu(myWnd), ID_SYSTEM_0+idx, MF_UNCHECKED);
						}
					} 
					// Special case for 99/4
					if (nSystem == 0) {
						keyboard=KEY_994;	// 99/4 layout
						MessageBox(myWnd, "Note: 99/4 does not always start. Exit and restart Classic99 to work around this. No PS/2 Keyboard support available.", "Compatibility Notes", MB_ICONINFORMATION|MB_OK);
					} else {
						// this lets the user disable ps2 keyboard support
						if (ps2keyboardok) {
							keyboard=KEY_994A_PS2;	// 99/4A with ps/2
						} else {
							keyboard=KEY_994A;		// 99/4A without ps/2
						}
					}
					SendMessage(hwnd, WM_COMMAND, ID_FILE_RESET, 0);
				}
			}
			if ((wParam >= ID_APP_0) && (wParam < ID_APP_0+100)) {
				// user requested to change cartridge (apps)
				int ret;
				if (fKeyEverPressed) {
					ret=MessageBox(hwnd, "This will reset the emulator - are you sure?", "Change Cartridge", MB_YESNO|MB_ICONQUESTION);
				} else {
					ret=IDYES;
				}
				if (IDYES == ret) {
					int idx;
					bUseSDCard=false;
					nCartGroup=0;
					nCart=wParam-ID_APP_0;
					for (idx=0; idx<100; idx++) {
						if (idx == nCart) {
							CheckMenuItem(GetMenu(myWnd), ID_APP_0+idx, MF_CHECKED);
						} else {
							CheckMenuItem(GetMenu(myWnd), ID_APP_0+idx, MF_UNCHECKED);
						}
					}
					for (idx=0; idx<100; idx++) {
						CheckMenuItem(GetMenu(myWnd), ID_GAME_0+idx, MF_UNCHECKED);
					}
					for (idx=0; idx<100; idx++) {
						CheckMenuItem(GetMenu(myWnd), ID_USER_0+idx, MF_UNCHECKED);
					}
					SendMessage(hwnd, WM_COMMAND, ID_FILE_RESET, 0);
				}
			}
			if ((wParam >= ID_GAME_0) && (wParam < ID_GAME_0+100)) {
				// user requested to change cartridge (games)
				int ret;
				if (fKeyEverPressed) {
					ret=MessageBox(hwnd, "This will reset the emulator - are you sure?", "Change Cartridge", MB_YESNO|MB_ICONQUESTION);
				} else {
					ret=IDYES;
				}
				if (IDYES == ret) {
					int idx;
					bUseSDCard=false;
					nCartGroup=1;
					nCart=wParam-ID_GAME_0;
					for (idx=0; idx<100; idx++) {
						if (idx == nCart) {
							CheckMenuItem(GetMenu(myWnd), ID_GAME_0+idx, MF_CHECKED);
						} else {
							CheckMenuItem(GetMenu(myWnd), ID_GAME_0+idx, MF_UNCHECKED);
						}
					}
					for (idx=0; idx<100; idx++) {
						CheckMenuItem(GetMenu(myWnd), ID_APP_0+idx, MF_UNCHECKED);
					}
					for (idx=0; idx<100; idx++) {
						CheckMenuItem(GetMenu(myWnd), ID_USER_0+idx, MF_UNCHECKED);
					}
					SendMessage(hwnd, WM_COMMAND, ID_FILE_RESET, 0);
				}
			}
			if ((wParam >= ID_USER_0) && (wParam < ID_USER_0+100)) {
				// user requested to change cartridge (user)
				int ret;
				if (fKeyEverPressed) {
					ret=MessageBox(hwnd, "This will reset the emulator - are you sure?", "Change Cartridge", MB_YESNO|MB_ICONQUESTION);
				} else {
					ret=IDYES;
				}
				if (IDYES == ret) {
					int idx;
					bUseSDCard=false;
					nCartGroup=2;
					nCart=wParam-ID_USER_0;
					for (idx=0; idx<100; idx++) {
						if (idx == nCart) {
							CheckMenuItem(GetMenu(myWnd), ID_USER_0+idx, MF_CHECKED);
						} else {
							CheckMenuItem(GetMenu(myWnd), ID_USER_0+idx, MF_UNCHECKED);
						}
					}
					for (idx=0; idx<100; idx++) {
						CheckMenuItem(GetMenu(myWnd), ID_GAME_0+idx, MF_UNCHECKED);
					}
					for (idx=0; idx<100; idx++) {
						CheckMenuItem(GetMenu(myWnd), ID_APP_0+idx, MF_UNCHECKED);
					}
					SendMessage(hwnd, WM_COMMAND, ID_FILE_RESET, 0);
				}
			}
			// Any others?
			switch (wParam)
			{
			case ID_SDGROM_SELECT:
			// activate or reset SDGROM support
			{
				int ret;
				if (fKeyEverPressed) {
					ret=MessageBox(hwnd, "This will reset the emulator - are you sure?", "Change Cartridge", MB_YESNO|MB_ICONQUESTION);
				} else {
					ret=IDYES;
				}
				if (IDYES == ret) {
					nCartGroup=3;
					nCart=0;
					bUseSDCard=true;
					SendMessage(hwnd, WM_COMMAND, ID_FILE_RESET, 0);
				}
			}
			break;

			case ID_CART_USER_OPEN:
			// Browse for user cartridge filename
			{
				OPENFILENAME ofn;
				char buf[256], buf2[256];

				memset(&ofn, 0, sizeof(OPENFILENAME));
				ofn.lStructSize=sizeof(OPENFILENAME);
				ofn.hwndOwner=hwnd;
				ofn.lpstrFilter="V9T9 Carts\0*.bin\0\0";
				strcpy(buf, "");
				ofn.lpstrFile=buf;
				ofn.nMaxFile=256;
				strcpy(buf2, "");
				ofn.lpstrFileTitle=buf2;
				ofn.nMaxFileTitle=256;
				ofn.Flags=OFN_HIDEREADONLY|OFN_FILEMUSTEXIST;
				
				char szTmpDir[MAX_PATH];
				GetCurrentDirectory(MAX_PATH, szTmpDir);

				if ((GetOpenFileName(&ofn))&&(ofn.nFileExtension > 1)) {
					int nCnt, nIdx;

					int nUsr=0;
					while (Users[nUsr].Img[0].nType != 0) {
						nUsr++;
						if (nUsr > 99) {
							nUsr=0;
							break;
						}
					}

					strncpy(Users[nUsr].szName, ofn.lpstrFileTitle, 32);
					Users[nUsr].szName[31]='\0';
					Users[nUsr].pDisk=NULL;
					Users[nUsr].szMessage=NULL;
					nCnt=0;
					for (nIdx=0; nIdx<4; nIdx++) {
						switch (nIdx) {
							case 0:	
								ofn.lpstrFile[ofn.nFileExtension-2]='C'; 
								Users[nUsr].Img[nCnt].nType=TYPE_ROM;
								break;
							case 1: 
								ofn.lpstrFile[ofn.nFileExtension-2]='D'; 
								Users[nUsr].Img[nCnt].nType=TYPE_XB;
								break;
							case 2: 
								ofn.lpstrFile[ofn.nFileExtension-2]='G'; 
								Users[nUsr].Img[nCnt].nType=TYPE_GROM;
								break;
							case 3: 
								ofn.lpstrFile[ofn.nFileExtension-2]='#'; 
								Users[nUsr].Img[nCnt].nType=TYPE_NONE;
								break;
						}
						if (nCnt < 3) {
							FILE *fp=fopen(ofn.lpstrFile, "rb");
							if (NULL != fp) {
								fseek(fp, 0, SEEK_END);
								Users[nUsr].Img[nCnt].dwImg=NULL;
								Users[nUsr].Img[nCnt].nLoadAddr=(unsigned short)0x6000;
								Users[nUsr].Img[nCnt].nLength=(unsigned short)ftell(fp);
								strncpy(Users[nUsr].Img[nCnt].szFileName, ofn.lpstrFile, 1024);
								Users[nUsr].Img[nCnt].szFileName[1023]='\0';
								nCnt++;
								fclose(fp);
								debug_write("Found %s...", ofn.lpstrFile);
							} else {
								Users[nUsr].Img[nCnt].nType=TYPE_NONE;
							}
						}
					}
					bUseSDCard=false;
					nCartGroup=2;
					nCart=nUsr;
					SendMessage(hwnd, WM_COMMAND, ID_FILE_RESET, 0);
				}
				SetCurrentDirectory(szTmpDir);
			}
			break;

			case ID_HELP_ABOUT:
				sprintf(szTemp, "Classic99 %s\n"\
								"(C)1994-2010\n\n"\
								"By Mike Brent (Tursi)\n"\
								"ROM data included under license from Texas Instruments.\n\n"\
								"Thanks to TI, Roland Meier, Jeff Brown, Frank\n"\
								"Palazolo, Ralph Nebet and Thierry Nouspikel for assistance.\n\n"\
								"Contains additional code by:\n"\
								"Joe Delekto - SAMS support\n"\
								"MESS Team - Speech, with thanks to\n"\
								"Ralph Nebet for speech ROM help.\n"\
								"Dag Lem and the reSID team\n"\
								"John Butler - 9900 Disasm\n"\
								"Derek Liauw Kie Fa - 2xSaI Renderer\n"\
								"2xSaI code from the SNES9x project\n\n"\
								"hq4x code by Maxim Stepin\n\n"\
								"Shay Green for the TV Filter\n\n"\
								"Keyboard map by Ron Reuter - www.mainbyte.com\n\n"\
								"tursi@harmlesslion.com\n"\
								"http://harmlesslion.com/software/classic99", 
						VERSION);
				MessageBox(myWnd, szTemp, "Classic99 About", MB_OK);
				break;

			case ID_HELP_KBMAP: 
				{
					if (NULL == hKBMap) {
						// create a modeless dialog to show the keyboard map
						hKBMap=CreateDialog(NULL, MAKEINTRESOURCE(IDD_KBMAP), hwnd, KBMapProc);
						ShowWindow(hKBMap, SW_SHOW);
					}
				}
				break;

			case ID_HELP_KNOWNISSUES:
				if (NULL == pCurrentHelpMsg) {
					pCurrentHelpMsg="There are no known issues with the currently selected cartridge.";
				}
				MessageBox(myWnd, pCurrentHelpMsg, "Compatibility Notes", MB_ICONINFORMATION|MB_OK);
				break;

			case ID_FILE_SCRAMBLERESET:
				bScrambleMemory=true;
				// fall through
			case ID_FILE_RESET:
				if (LOWORD(wParam) == ID_FILE_RESET) {
					bScrambleMemory=false;
				}
				
				TriggerBreakPoint();					// halt the CPU
				Sleep(50);								// wait for it...

				ST=(ST&0xfff0);							// disable interrupts
				memset(CRU, 1, 4096);					// reset 9901
				CRU[0]=0;								// except not in timer mode
				timer9901=0;
				timer9901IntReq=0;
				starttimer9901=0;
				timer9901IntOk=0;
				wrword(0x83c4,0);						// Console bug work around, make sure no user int is active
				init_kb();								// Reset keyboard emulation
				SetupSams(sams_enabled, sams_size);		// Prepare the AMS system
				if (NULL != InitSid) {
					InitSid();							// reset the SID chip
					if (NULL != SetSidBanked) {		
						SetSidBanked(false);			// switch it out for now
					}
				}
				readroms();								// reload the real ROMs
				if (NULL != pCurrentHelpMsg) {
					szDefaultWindowText="Classic99 - See Help->Known Issues for this cart";
					SetWindowText(myWnd, szDefaultWindowText);
				} else {
					szDefaultWindowText="Classic99";
					SetWindowText(myWnd, szDefaultWindowText);
				}

				WP=romword(0x0000);						// Reset 
				PC=romword(0x0002);
				for (int idx=0; idx<=PCODEGROMBASE; idx++) {
					GROMBase[idx].grmaccess=2;			// no GROM accesses yet
				}
				nCurrentDSR=-1;
				memset(nDSRBank, 0, sizeof(nDSRBank));
				vdpaccess=0;							// No VDP address writes yet 
				vdpprefetch=0;
				tmpVDPADD=0;							// part of the vdpaccess
				interrupt_needed=0;						// No interrupt missed yet
				CPUSpeechHalt=false;					// not halted for speech reasons
				CPUSpeechHaltByte=0;					// doesn't matter
				cpucount=0;
				cpuframes=0;
				nCycleCount=26;							// not that it's a big deal, but that's how long reset takes ;)
				fKeyEverPressed=false;					// No key pressed yet (to disable the warning on cart change)
				DoPlay();
				cycles_left=max_cpf=(hzRate==50?DEFAULT_50HZ_CPF:DEFAULT_60HZ_CPF);
				InterlockedExchange((LONG*)&cycles_left, max_cpf);
				break;

			case ID_FILE_QUIT:
				if (IDYES == MessageBox(myWnd, "Are you sure you want to quit?", "Classic99 go ByeBye?", MB_YESNO)) {
					MuteAudio();
					quitflag=1;
					PostQuitMessage(0);
				} 
				break;

			case ID_VIDEO_FLICKER:
				if (lParam != 1) {
					bUse5SpriteLimit=!bUse5SpriteLimit;
				}

				if (!bUse5SpriteLimit) {
					CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FLICKER, MF_UNCHECKED);
				} else {
					CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FLICKER, MF_CHECKED);
				}
				break;

			case ID_LAYERS_DISABLEBLANKING:
				if (bDisableBlank) {
					bDisableBlank=false;
					CheckMenuItem(GetMenu(myWnd), ID_LAYERS_DISABLEBLANKING, MF_UNCHECKED);
				} else {
					bDisableBlank=true;
					CheckMenuItem(GetMenu(myWnd), ID_LAYERS_DISABLEBLANKING, MF_CHECKED);
				} 
				SetEvent(DisplayEvent);
				break;

			case ID_LAYERS_DISABLESPRITES:
				if (bDisableSprite) {
					bDisableSprite=false;
					CheckMenuItem(GetMenu(myWnd), ID_LAYERS_DISABLESPRITES, MF_UNCHECKED);
				} else {
					bDisableSprite=true;
					CheckMenuItem(GetMenu(myWnd), ID_LAYERS_DISABLESPRITES, MF_CHECKED);
				} 
				SetEvent(DisplayEvent);
				break;

			case ID_LAYERS_DISABLEBACKGROUND:
				if (bDisableBackground) {
					bDisableBackground=false;
					CheckMenuItem(GetMenu(myWnd), ID_LAYERS_DISABLEBACKGROUND, MF_UNCHECKED);
				} else {
					bDisableBackground=true;
					CheckMenuItem(GetMenu(myWnd), ID_LAYERS_DISABLEBACKGROUND, MF_CHECKED);
				} 
				SetEvent(DisplayEvent);
				break;
			
			case ID_VIDEO_50HZ:
				if (lParam != 1) {
					hzRate=(hzRate==60)?50:60;
				}

				if (hzRate == 60) {
					CheckMenuItem(GetMenu(myWnd), ID_VIDEO_50HZ, MF_UNCHECKED);
				} else {
					CheckMenuItem(GetMenu(myWnd), ID_VIDEO_50HZ, MF_CHECKED);
				}
				break;

			case ID_VIDEO_STARTRECORDING:
				if (Recording == 0)
				{
					debug_write("Starting AVI recording");
					if (0==InitAvi(false))
					{
						szDefaultWindowText="Classic99 - Recording AVI";
						SetWindowText(myWnd, szDefaultWindowText);
						Recording=1;
						RecordFrame=0;
					}
				}
				break;

			case ID_VIDEO_STARTRECORDINGVIDEO:
				// with audio
				if (Recording == 0)
				{
					debug_write("Starting AVI recording with at audio %dHz", AudioSampleRate);
					if (0==InitAvi(true))
					{
						szDefaultWindowText="Classic99 - Recording AVI";
						SetWindowText(myWnd, szDefaultWindowText);
						Recording=1;
						RecordFrame=0;
					}
				}
				break;
			
			case ID_VIDEO_STOPRECORDING:
				if (Recording)
				{
					debug_write("Stoping AVI recording");
					szDefaultWindowText="Classic99";
					SetWindowText(myWnd, szDefaultWindowText);
					CloseAVI();
					Recording=0;
				}
				break;

			case ID_VIDEO_SCREENSHOTBAS:
				// save the current raw TI image
				SaveScreenshot(false, false);
				break;
			
			case ID_VIDEO_SCREENSHOTFILT:
				// not supported yet
				SaveScreenshot(false, true);
				break;

			case ID_VIDEO_MAINTAINASPECT:
				if (1 != lParam) {
					MaintainAspect=MaintainAspect?0:1;
				}
				if (MaintainAspect)
				{
					CheckMenuItem(GetMenu(myWnd), ID_VIDEO_MAINTAINASPECT, MF_CHECKED);
				}
				else
				{
					CheckMenuItem(GetMenu(myWnd), ID_VIDEO_MAINTAINASPECT, MF_UNCHECKED);
				}
				break;

			case ID_OPTIONS_PAUSEINACTIVE:
				if (1 != lParam) {
					PauseInactive=PauseInactive?0:1;
				}
				if (PauseInactive) {
					CheckMenuItem(GetMenu(myWnd), ID_OPTIONS_PAUSEINACTIVE, MF_CHECKED);
				} else {
					CheckMenuItem(GetMenu(myWnd), ID_OPTIONS_PAUSEINACTIVE, MF_UNCHECKED);
				}
				break;

			case ID_OPTIONS_SPEECHENABLED:
				if (1 != lParam) {
					SpeechEnabled=SpeechEnabled?0:1;
				}
				if (SpeechEnabled) {
					CheckMenuItem(GetMenu(myWnd), ID_OPTIONS_SPEECHENABLED, MF_CHECKED);
				} else {
					CheckMenuItem(GetMenu(myWnd), ID_OPTIONS_SPEECHENABLED, MF_UNCHECKED);
				}
				break;


			case ID_OPTIONS_CPUTHROTTLING:
				{
					// Make sure nothing's left over from the old speed
					InterlockedExchange((LONG*)&cycles_left, 0);

					// Now just used to set the check boxes correctly
					if ((CPUThrottle == CPU_NORMAL) && (SystemThrottle == VDP_CPUSYNC) && (max_cpf > SLOW_CPF)) {
						CheckMenuItem(GetMenu(myWnd), ID_CPUTHROTTLING_NORMAL, MF_CHECKED);
						szDefaultWindowText="Classic99";
					} else {
						CheckMenuItem(GetMenu(myWnd), ID_CPUTHROTTLING_NORMAL, MF_UNCHECKED);
					}

					if ((CPUThrottle == CPU_NORMAL) && (SystemThrottle == VDP_CPUSYNC) && (max_cpf == SLOW_CPF)) {
						CheckMenuItem(GetMenu(myWnd), ID_CPUTHROTTLING_CPUSLOW, MF_CHECKED);
						szDefaultWindowText="Classic99 - Slow CPU";
					} else {
						CheckMenuItem(GetMenu(myWnd), ID_CPUTHROTTLING_CPUSLOW, MF_UNCHECKED);
					}

					if ((CPUThrottle == CPU_OVERDRIVE) && (SystemThrottle == VDP_REALTIME)) {
						CheckMenuItem(GetMenu(myWnd), ID_CPUTHROTTLING_CPUOVERDRIVE, MF_CHECKED);
						szDefaultWindowText="Classic99 - CPU Overdrive";
					} else {
						CheckMenuItem(GetMenu(myWnd), ID_CPUTHROTTLING_CPUOVERDRIVE, MF_UNCHECKED);
					}

					if ((CPUThrottle == CPU_MAXIMUM) && (SystemThrottle == VDP_CPUSYNC)) {
						CheckMenuItem(GetMenu(myWnd), ID_CPUTHROTTLING_SYSTEMMAXIMUM, MF_CHECKED);
						szDefaultWindowText="Classic99 - System Maximum";
					} else {
						CheckMenuItem(GetMenu(myWnd), ID_CPUTHROTTLING_SYSTEMMAXIMUM, MF_UNCHECKED);
					}
					SetWindowText(myWnd, szDefaultWindowText);
				}
				break;

			case ID_CPUTHROTTLING_NORMAL:
				CPUThrottle=CPU_NORMAL;
				SystemThrottle=VDP_CPUSYNC;
				max_cpf=oldmax;
				if (max_cpf <= SLOW_CPF) {
					// we've lost the configured speed.. sorry? to remove that later anyway
					max_cpf=(hzRate==50?DEFAULT_50HZ_CPF:DEFAULT_60HZ_CPF);
					oldmax=(hzRate==50?DEFAULT_50HZ_CPF:DEFAULT_60HZ_CPF);
				}
				PostMessage(myWnd, WM_COMMAND, ID_OPTIONS_CPUTHROTTLING, 1);
				break;

			case ID_CPUTHROTTLING_CPUSLOW:
				CPUThrottle=CPU_NORMAL;
				SystemThrottle=VDP_CPUSYNC;
				oldmax=max_cpf;
				max_cpf=SLOW_CPF;
				PostMessage(myWnd, WM_COMMAND, ID_OPTIONS_CPUTHROTTLING, 1);
				break;

			case ID_CPUTHROTTLING_CPUOVERDRIVE:
				CPUThrottle=CPU_OVERDRIVE;
				SystemThrottle=VDP_REALTIME;
				max_cpf=oldmax;
				PostMessage(myWnd, WM_COMMAND, ID_OPTIONS_CPUTHROTTLING, 1);
				break;

			case ID_CPUTHROTTLING_SYSTEMMAXIMUM:
				CPUThrottle=CPU_MAXIMUM;
				SystemThrottle=VDP_CPUSYNC;
				max_cpf=oldmax;
				PostMessage(myWnd, WM_COMMAND, ID_OPTIONS_CPUTHROTTLING, 1);
				break;

			case ID_OPTIONS_AUDIO: 
				{
					if (Recording) {
						MessageBox(myWnd, "Can't change audio options while recording video", "Classic99 Error", MB_OK);
						break;
					}

					// Create a dialog to reconfigure audio voices
					if (DialogBox(NULL, MAKEINTRESOURCE(IDD_AUDIO), hwnd, AudioBoxProc) == IDOK) {
						DoPause();					// stop the system
						Sleep(100);					// wait for it
						// now reset audio
						sound_init(AudioSampleRate);
						GenerateToneBuffer();
						GenerateSIDBuffer();
						// now continue
						DoPlay();
					}
				}
				break;

			case ID_OPTIONS_OPTIONS:
				{
					// Create a dialog to reconfigure generic options
					DialogBox(NULL, MAKEINTRESOURCE(IDD_OPTIONS), hwnd, OptionsBoxProc);
					// It handles the OK/Cancel operation in OptionsBoxProc
				}
				break;

			case ID_OPTIONS_TV:
				// If filters aren't available, tell the user
				if (!TVFiltersAvailable) {
					MessageBox(hwnd, "TV Filter DLL is not available.", "Classic99", MB_OK);
					break;
				}
				// If we aren't in TV mode, switch to it and come back here
				if (4 != FilterMode) {
					PostMessage(hwnd, WM_COMMAND, ID_VIDEO_FILTERMODE_TVMODE, 0);
					PostMessage(hwnd, WM_COMMAND, ID_OPTIONS_TV, 0);
					break;
				}
				// If the TV dialog doesn't already exist, exist it :)
				if (NULL == hTVDlg) {
					// create a modeless dialog to show the TV controls dialog
					hTVDlg=CreateDialog(NULL, MAKEINTRESOURCE(IDD_TVOPTIONS), hwnd, TVBoxProc);
					ShowWindow(hTVDlg, SW_SHOW);
				}
				break;

			case ID_EDITPASTE:
				if (NULL != PasteString) {
					Beep(550,100);
					break;
				}
				if (OpenClipboard(myWnd)) {
					HANDLE data;

					data=GetClipboardData(CF_TEXT);

					if (data) {
						PasteString=(char*)malloc(strlen((char*)data)+1);
						strcpy(PasteString, (char*)data);
						PasteIndex=PasteString;
					}

					CloseClipboard();
				}
				break;
			
			case ID_EDIT_DEBUGGER:
				// activate debugger screen
				if (NULL == dbgWnd) {
					LaunchDebugWindow();
				}
				break;

			case ID_EDIT_BUG99WINDOW:
				ShowWindow(hBugWnd, SW_SHOW);
				break;

			case ID_EDIT_HEATMAP: 
				// activate heatmap screen
				if (NULL == hHeatMap) {
					// create a modeless dialog to show the keyboard map
					hHeatMap=CreateDialog(NULL, MAKEINTRESOURCE(IDD_HEATMAP), hwnd, HeatMapProc);
					ShowWindow(hHeatMap, SW_SHOW);
				}
				break;

			case ID_CHANGESIZE_1X:
				{
					GetWindowRect(myWnd, &myrect);
					GetClientRect(myWnd, &myrect2);
					myrect.right = myrect.left + CurrentDDSD.dwWidth + ((myrect.right - myrect.left)-(myrect2.right - myrect2.left));
					myrect.bottom = myrect.top + CurrentDDSD.dwHeight + ((myrect.bottom - myrect.top)-(myrect2.bottom - myrect2.top));
					MoveWindow(myWnd, myrect.left, myrect.top, myrect.right-myrect.left, myrect.bottom-myrect.top, true);
				}

				if (0 == StretchMode) {
					GetClientRect(myWnd, &myrect);
					myDC=GetDC(myWnd);
					FillRect(myDC, &myrect, (HBRUSH)(COLOR_MENU+1));
					ReleaseDC(myWnd, myDC);
				}

				GetClientRect(myWnd, &myrect2);
				debug_write("Client rect now %d x %d", myrect2.right-myrect2.left, myrect2.bottom-myrect2.top);

				// repeat it once only - this accounts for the menu changing from 1 to 2 lines and back
				if (lParam != 99) {
					if ((myrect2.right-myrect2.left != CurrentDDSD.dwWidth) || (myrect2.bottom-myrect2.top != CurrentDDSD.dwHeight)) {
						PostMessage(myWnd, WM_COMMAND, ID_CHANGESIZE_1X, 99);
					}
				}
				break;

			case ID_CHANGESIZE_2X:
				{
					GetWindowRect(myWnd, &myrect);
					GetClientRect(myWnd, &myrect2);
					myrect.right = myrect.left + CurrentDDSD.dwWidth*2 + ((myrect.right - myrect.left)-(myrect2.right - myrect2.left));
					myrect.bottom = myrect.top + CurrentDDSD.dwHeight*2 + ((myrect.bottom - myrect.top)-(myrect2.bottom - myrect2.top));
					MoveWindow(myWnd, myrect.left, myrect.top, myrect.right-myrect.left, myrect.bottom-myrect.top, true);
				}

				if (0 == StretchMode) {
					GetClientRect(myWnd, &myrect);
					myDC=GetDC(myWnd);
					FillRect(myDC, &myrect, (HBRUSH)(COLOR_MENU+1));
					ReleaseDC(myWnd, myDC);
				}

				GetClientRect(myWnd, &myrect2);
				debug_write("Client rect now %d x %d", myrect2.right-myrect2.left, myrect2.bottom-myrect2.top);

				// repeat it once only - this accounts for the menu changing from 1 to 2 lines and back
				if (lParam != 99) {
					if ((myrect2.right-myrect2.left != CurrentDDSD.dwWidth*2) || (myrect2.bottom-myrect2.top != CurrentDDSD.dwHeight*2)) {
						PostMessage(myWnd, WM_COMMAND, ID_CHANGESIZE_2X, 99);
					}
				}
				break;

			case ID_CHANGESIZE_4X:
				{
					GetWindowRect(myWnd, &myrect);
					GetClientRect(myWnd, &myrect2);
					myrect.right = myrect.left + CurrentDDSD.dwWidth*4 + ((myrect.right - myrect.left)-(myrect2.right - myrect2.left));
					myrect.bottom = myrect.top + CurrentDDSD.dwHeight*4 + ((myrect.bottom - myrect.top)-(myrect2.bottom - myrect2.top));
					MoveWindow(myWnd, myrect.left, myrect.top, myrect.right-myrect.left, myrect.bottom-myrect.top, true);
				}

				if (0 == StretchMode) {
					GetClientRect(myWnd, &myrect);
					myDC=GetDC(myWnd);
					FillRect(myDC, &myrect, (HBRUSH)(COLOR_MENU+1));
					ReleaseDC(myWnd, myDC);
				}

				GetClientRect(myWnd, &myrect2);
				debug_write("Client rect now %d x %d", myrect2.right-myrect2.left, myrect2.bottom-myrect2.top);

				// repeat it once only - this accounts for the menu changing from 1 to 2 lines and back
				if (lParam != 99) {
					if ((myrect2.right-myrect2.left != CurrentDDSD.dwWidth*4) || (myrect2.bottom-myrect2.top != CurrentDDSD.dwHeight*4)) {
						PostMessage(myWnd, WM_COMMAND, ID_CHANGESIZE_4X, 99);
					}
				}
				break;

			case ID_VIDEO_STRETCHMODE_NONE:		// This mode is a fallback, it must not fail and must not loop back
				StretchMode=0;
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_STRETCHMODE_NONE, MF_CHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_STRETCHMODE_DIB, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_STRETCHMODE_DX, MF_UNCHECKED);
				InvalidateRect(myWnd, NULL, false);
				break;

			case ID_VIDEO_STRETCHMODE_DIB:
				StretchMode=1;
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_STRETCHMODE_NONE, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_STRETCHMODE_DIB, MF_CHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_STRETCHMODE_DX, MF_UNCHECKED);
				InvalidateRect(myWnd, NULL, false);
				break;

			case ID_VIDEO_STRETCHMODE_DX:
				StretchMode=2;
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_STRETCHMODE_NONE, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_STRETCHMODE_DIB, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_STRETCHMODE_DX, MF_CHECKED);
				takedownDirectDraw();
				SetupDirectDraw(0);
				if (2 != StretchMode) {
					myproc(hwnd, WM_COMMAND, ID_VIDEO_STRETCHMODE_NONE, 0);
				} else {
					InvalidateRect(myWnd, NULL, false);
				}
				break;

			case ID_VIDEO_STRETCHMODE_DXFULL_320X240X8:
			case ID_VIDEO_STRETCHMODE_DXFULL_640X480X8:
			case ID_VIDEO_STRETCHMODE_DXFULL_640X480X16:
			case ID_DXFULL_640X480X32:
			case ID_VIDEO_STRETCHMODE_DXFULL_800X600X16:
			case ID_DXFULL_800X600X32:
			case ID_DXFULL_1024X768X16:
			case ID_DXFULL_1024X768X32:
				StretchMode=3;
				FullScreenMode=wParam-ID_VIDEO_STRETCHMODE_DXFULL_320X240X8+1;
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_STRETCHMODE_NONE, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_STRETCHMODE_DIB, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_STRETCHMODE_DX, MF_UNCHECKED);
				{
					for (DWORD idx=ID_VIDEO_STRETCHMODE_DXFULL_320X240X8; idx<ID_DXFULL_1024X768X32; idx++) {
						if (idx==wParam) {
							CheckMenuItem(GetMenu(myWnd), wParam, MF_CHECKED);
						} else {
							CheckMenuItem(GetMenu(myWnd), wParam, MF_UNCHECKED);
						}
					}
				}
				takedownDirectDraw();
				SetupDirectDraw(FullScreenMode);
				if (3 != StretchMode) {
					myproc(hwnd, WM_COMMAND, ID_VIDEO_STRETCHMODE_NONE, 0);
				} else {
					InvalidateRect(myWnd, NULL, false);
				}
				break;

			case ID_VIDEO_FILTERMODE_NONE:
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_NONE, MF_CHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_HQ4X, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_2XSAI, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_SUPER2XSAI, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_SUPEREAGLE, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_TVMODE, MF_UNCHECKED);
				if (FilterMode != 0) {
					ResizeBackBuffer(256+16, 192+16);
				}
				FilterMode=0;
				InvalidateRect(myWnd, NULL, false);
				break;

			case ID_VIDEO_FILTERMODE_HQ4X:
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_NONE, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_HQ4X, MF_CHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_2XSAI, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_SUPER2XSAI, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_SUPEREAGLE, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_TVMODE, MF_UNCHECKED);
				if (FilterMode != 5) {
					ResizeBackBuffer((256+16)*4, (192+16)*4);
				}
				FilterMode=5;
				InvalidateRect(myWnd, NULL, false);
				break;

			case ID_VIDEO_FILTERMODE_2XSAI:
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_NONE, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_2XSAI, MF_CHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_HQ4X, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_SUPER2XSAI, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_SUPEREAGLE, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_TVMODE, MF_UNCHECKED);
				if (FilterMode != 1) {
					ResizeBackBuffer(512+32,384+29);
				}
				FilterMode=1;
				InvalidateRect(myWnd, NULL, false);
				break;

			case ID_VIDEO_FILTERMODE_SUPER2XSAI:
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_NONE, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_2XSAI, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_SUPER2XSAI, MF_CHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_HQ4X, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_SUPEREAGLE, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_TVMODE, MF_UNCHECKED);
				if (FilterMode != 2) {
					ResizeBackBuffer(512+32,384+29);
				}
				FilterMode=2;
				InvalidateRect(myWnd, NULL, false);
				break;

			case ID_VIDEO_FILTERMODE_SUPEREAGLE:
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_NONE, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_HQ4X, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_2XSAI, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_SUPER2XSAI, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_SUPEREAGLE, MF_CHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_TVMODE, MF_UNCHECKED);
				if (FilterMode != 3) {
					ResizeBackBuffer(512+32,384+29);
				}
				FilterMode=3;
				InvalidateRect(myWnd, NULL, false);
				break;

			case ID_VIDEO_FILTERMODE_TVMODE:
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_NONE, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_HQ4X, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_2XSAI, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_SUPER2XSAI, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_SUPEREAGLE, MF_UNCHECKED);
				CheckMenuItem(GetMenu(myWnd), ID_VIDEO_FILTERMODE_TVMODE, MF_CHECKED);
				if (FilterMode != 4) {
					ResizeBackBuffer(640,384+29);
				}
				FilterMode=4;
				InvalidateRect(myWnd, NULL, false);
				break;

			default:
				SetSoundVolumes();
				return(DefWindowProc(hwnd, msg, wParam, lParam));
			}
			SetSoundVolumes();
			break;

		case WM_SIZE: 
			{
				HDC myDC;
				
//				MuteAudio();

				if (MaintainAspect)
				{
					GetWindowRect(myWnd, &myrect);
					GetClientRect(myWnd, &myrect2);
					i1=((myrect.right - myrect.left)-(myrect2.right - myrect2.left))/2;
					myrect.left += i1;
					myrect.right = myrect.left + myrect2.right;
					myrect.bottom -= i1;
					myrect.top = myrect.bottom - myrect2.bottom;

					double ratio=(double)CurrentDDSD.dwHeight / (double)CurrentDDSD.dwWidth;
					
					height=(float)((myrect.right - myrect.left) * ratio);
					height-=(myrect.bottom - myrect.top);

					if (height)
					{
						GetWindowRect(myWnd, &myrect);
						myrect.bottom+=(long)height;
						MoveWindow(myWnd, myrect.left, myrect.top, myrect.right-myrect.left, myrect.bottom-myrect.top, true);
					}
				}

				if (0 == StretchMode) {
					GetClientRect(myWnd, &myrect);
					myDC=GetDC(myWnd);
					FillRect(myDC, &myrect, (HBRUSH)(COLOR_MENU+1));
					ReleaseDC(myWnd, myDC);
				}
				SetSoundVolumes();
			}
			break;

		case WM_MOUSEACTIVATE:
			SetEvent(BlitEvent);
			return(DefWindowProc(hwnd, msg, wParam, lParam));
			break;

		case WM_RBUTTONUP:	// exit full screen
			if (3 == StretchMode) {
				MuteAudio();
				StretchMode=2;
				takedownDirectDraw();
				SetupDirectDraw(false);
				SetSoundVolumes();
			}
			break;

		case WM_SETFOCUS: 
			if (PauseInactive) {
				// Re-enable sounds
				SetSoundVolumes();
			}
			break;

		case WM_KILLFOCUS:
			// Disable sounds, cache current levels
			if (PauseInactive) {
				MuteAudio();
			}
			// clear all keyboard state
			memset(key, 0, sizeof(key));
			init_kb();
			break;

		default:
			return(DefWindowProc(hwnd, msg, wParam, lParam));
		}
		return 0;
	}

	return(DefWindowProc(hwnd, msg, wParam, lParam));
}


BOOL CALLBACK AudioBoxProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	static int nRate=22050;
	static int nLocalSidEnable = false;

    switch (uMsg) 
    { 
        case WM_COMMAND: 
            switch (LOWORD(wParam)) 
            { 
				case IDC_11KHZ:
					nRate=11025;
					break;
				case IDC_22KHZ:
					nRate=22050;
					break;
				case IDC_44KHZ:
					nRate=44100;
					break;

				case IDC_SIDENABLE:
					nLocalSidEnable = (SendDlgItemMessage(hwnd, IDC_SIDENABLE, BM_GETCHECK, 0, 0) == BST_CHECKED);
					break;

                case IDOK: 
					max_volume=SendDlgItemMessage(hwnd, IDC_SLDVOL, TBM_GETPOS, 0, 0);
					AudioSampleRate=nRate;
					if (NULL == SetSidEnable) {
						if (nLocalSidEnable) {
							MessageBox(myWnd, "SID DLL is not available", "Classic99 Support File Missing", MB_OK);
						}
					} else {
						SetSidEnable(nLocalSidEnable != 0);
					}
                    // Fall through. 
                 case IDCANCEL: 
                    EndDialog(hwnd, wParam); 
                    return TRUE; 

				 case IDC_DEFAULT:
					// set the values to default
					SendDlgItemMessage(hwnd, IDC_SLDVOL, TBM_SETPOS, TRUE, 80);
					return TRUE;
            } 
			break;

		case WM_INITDIALOG:
			SendDlgItemMessage(hwnd, IDC_SLDVOL, TBM_SETRANGE, TRUE, MAKELONG(0,100));
			SendDlgItemMessage(hwnd, IDC_SLDVOL, TBM_SETPOS, TRUE, max_volume);
			SendDlgItemMessage(hwnd, IDC_11KHZ, BM_SETCHECK, BST_UNCHECKED, 0);
			SendDlgItemMessage(hwnd, IDC_22KHZ, BM_SETCHECK, BST_UNCHECKED, 0);
			SendDlgItemMessage(hwnd, IDC_44KHZ, BM_SETCHECK, BST_UNCHECKED, 0);
			switch (AudioSampleRate) {
				case 11025:
					nRate=11025;
					SendDlgItemMessage(hwnd, IDC_11KHZ, BM_SETCHECK, BST_CHECKED, 0);
					break;
				case 44100:
					nRate=44100;
					SendDlgItemMessage(hwnd, IDC_44KHZ, BM_SETCHECK, BST_CHECKED, 0);
					break;
				default:
					nRate=22050;
					SendDlgItemMessage(hwnd, IDC_22KHZ, BM_SETCHECK, BST_CHECKED, 0);
					break;
			}
			if (NULL != GetSidEnable) {
				nLocalSidEnable = GetSidEnable();
			} else {
				nLocalSidEnable = false;
			}
			SendDlgItemMessage(hwnd, IDC_SIDENABLE, BM_SETCHECK, nLocalSidEnable ? BST_CHECKED:BST_UNCHECKED, 0);
			return TRUE;
    } 
    return FALSE; 
} 

BOOL CALLBACK OptionsBoxProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
    switch (uMsg) 
    { 
        case WM_COMMAND: 
            switch (LOWORD(wParam)) 
            { 
                case IDOK: 
					// Read new values from dialog
					SendDlgItemMessage(hwnd, IDC_AVIFILENAME, WM_GETTEXT, 256, (LPARAM)AVIFileName);
					max_cpf=(hzRate==50?DEFAULT_50HZ_CPF:DEFAULT_60HZ_CPF)*SendDlgItemMessage(hwnd, IDC_SLDCPU, TBM_GETPOS, 0, 0)/10;
					drawspeed=SendDlgItemMessage(hwnd, IDC_SLDFRAMESKIP, TBM_GETPOS, 0, 0);
					slowdown_keyboard=IsDlgButtonChecked(hwnd, IDC_CHKSLOWKEY)?1:0;
					
					if (IsDlgButtonChecked(hwnd, IDC_THROTTLECPU)) {
						CPUThrottle=CPU_NORMAL;
						SystemThrottle=VDP_CPUSYNC;
					}
					if (IsDlgButtonChecked(hwnd, IDC_UNTHROTTLECPU)) {
						CPUThrottle=CPU_OVERDRIVE;
						SystemThrottle=VDP_REALTIME;
					}
					if (IsDlgButtonChecked(hwnd, IDC_UNTHROTTLEALL)) {
						CPUThrottle=CPU_MAXIMUM;
						SystemThrottle=VDP_CPUSYNC;
					}
					{
						// SAMS
						int old_sams_enabled=sams_enabled;
						int old_sams_size=sams_size;
						if (IsDlgButtonChecked(hwnd, IDC_AMS_0K)) {
							sams_enabled=0;
						}
						if (IsDlgButtonChecked(hwnd, IDC_AMS_128K)) {
							sams_enabled=1;
							sams_size=0;
						}
						if (IsDlgButtonChecked(hwnd, IDC_AMS_256K)) {
							sams_enabled=1;
							sams_size=1;
						}
						if (IsDlgButtonChecked(hwnd, IDC_AMS_512K)) {
							sams_enabled=1;
							sams_size=2;
						}
						if (IsDlgButtonChecked(hwnd, IDC_AMS_1024K)) {
							sams_enabled=1;
							sams_size=3;
						}
						if ((sams_enabled!=old_sams_enabled)||(sams_size!=old_sams_size)) {
							// changing this in real time wipes memory, so don't do that!
							int ret;
							ret=MessageBox(hwnd, "This will reset the emulator - are you sure?", "Change AMS Memory size", MB_YESNO|MB_ICONQUESTION);
							if (IDYES == ret) {
								PostMessage(myWnd, WM_COMMAND, ID_FILE_RESET, 0);
							} else {
								sams_enabled=old_sams_enabled;
								sams_size=old_sams_size;
							}
						}
					}

					// joysticks
					fJoy=IsDlgButtonChecked(hwnd, IDC_CHKJOYST)?1:0;
					if (IsDlgButtonChecked(hwnd, IDC_JOY1KEY))  joy1mode=0;
					if (IsDlgButtonChecked(hwnd, IDC_JOY1JOY1)) joy1mode=1;
					if (IsDlgButtonChecked(hwnd, IDC_JOY1JOY2)) joy1mode=2;
					if (IsDlgButtonChecked(hwnd, IDC_JOY2KEY))  joy2mode=0;
					if (IsDlgButtonChecked(hwnd, IDC_JOY2JOY1)) joy2mode=1;
					if (IsDlgButtonChecked(hwnd, IDC_JOY2JOY2)) joy2mode=2;
					fJoystickActiveOnKeys=0;		// reset here too

					// Special - tell the CPU Throttle menu item the new state
					PostMessage(myWnd, WM_COMMAND, ID_OPTIONS_CPUTHROTTLING, 1);
                    // Fall through. 
                 case IDCANCEL: 
                    EndDialog(hwnd, wParam); 
                    return TRUE; 

				 case IDC_BROWSEAVI:
					 // Browse for new AVI filename
					 {
						OPENFILENAME ofn;
						char buf[256];

						memset(&ofn, 0, sizeof(OPENFILENAME));
						ofn.lStructSize=sizeof(OPENFILENAME);
						ofn.hwndOwner=hwnd;
						ofn.lpstrFilter="AVI Files\0*.avi\0\0";
						SendDlgItemMessage(hwnd, IDC_AVIFILENAME, WM_GETTEXT, 256, (LPARAM)buf);
						ofn.lpstrFile=buf;
						ofn.nMaxFile=256;
						ofn.Flags=OFN_HIDEREADONLY|OFN_OVERWRITEPROMPT|OFN_PATHMUSTEXIST;

						char szTmpDir[MAX_PATH];
						GetCurrentDirectory(MAX_PATH, szTmpDir);

						if (GetSaveFileName(&ofn)) {
							SendDlgItemMessage(hwnd, IDC_AVIFILENAME, WM_SETTEXT, 0, (LPARAM)ofn.lpstrFile);
						}

						SetCurrentDirectory(szTmpDir);
					 }

					 return TRUE;
            } 
			break;

		case WM_INITDIALOG:
			// load the controls with current values
			SendDlgItemMessage(hwnd, IDC_AVIFILENAME, WM_SETTEXT, 0, (LPARAM)AVIFileName);
			SendDlgItemMessage(hwnd, IDC_SLDCPU, TBM_SETRANGE, TRUE, MAKELONG(1,20));
			SendDlgItemMessage(hwnd, IDC_SLDCPU, TBM_SETPOS, TRUE, (max_cpf*10)/(hzRate==50?DEFAULT_50HZ_CPF:DEFAULT_60HZ_CPF));
			SendDlgItemMessage(hwnd, IDC_SLDFRAMESKIP, TBM_SETRANGE, TRUE, MAKELONG(0,10));
			SendDlgItemMessage(hwnd, IDC_SLDFRAMESKIP, TBM_SETPOS, TRUE, drawspeed);
			SendDlgItemMessage(hwnd, IDC_CHKJOYST, BM_SETCHECK, fJoy?BST_CHECKED:BST_UNCHECKED, 0);
			SendDlgItemMessage(hwnd, IDC_CHKSLOWKEY, BM_SETCHECK, slowdown_keyboard?BST_CHECKED:BST_UNCHECKED, 0);
			
			// default
			CheckRadioButton(hwnd, IDC_THROTTLECPU, IDC_UNTHROTTLEALL, IDC_THROTTLECPU);
			if ((CPUThrottle == CPU_OVERDRIVE) && (SystemThrottle == VDP_REALTIME)) {
				CheckRadioButton(hwnd, IDC_THROTTLECPU, IDC_UNTHROTTLEALL, IDC_UNTHROTTLECPU);
			}
			if ((CPUThrottle == CPU_MAXIMUM) && (SystemThrottle == VDP_CPUSYNC)) {
				CheckRadioButton(hwnd, IDC_THROTTLECPU, IDC_UNTHROTTLEALL, IDC_UNTHROTTLEALL);
			}

			CheckRadioButton(hwnd, IDC_JOY1KEY, IDC_JOY1JOY2, IDC_JOY1KEY+joy1mode);
			CheckRadioButton(hwnd, IDC_JOY2KEY, IDC_JOY2JOY2, IDC_JOY2KEY+joy2mode);
			CheckRadioButton(hwnd, IDC_AMS_0K, IDC_AMS_1024K, sams_enabled?sams_size+IDC_AMS_128K:IDC_AMS_0K);
			return TRUE;

    } 
    return FALSE; 
} 

BOOL CALLBACK KBMapProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	// only tricky part is loading the kb image resource
	static HBITMAP hBmp=NULL;
	HWND hWnd=NULL;

    switch (uMsg) 
    { 
        case WM_COMMAND: 
            switch (LOWORD(wParam)) 
            { 
                case IDOK: 
                    // Fall through. 
                 case IDCANCEL: 
                    EndDialog(hwnd, wParam); 
					if (NULL != hBmp) {
						DeleteObject(hBmp);
						hBmp=NULL;
					}
					hKBMap=NULL;
                    return TRUE; 
            } 
			break;

		case WM_INITDIALOG:
			// load the bitmap into the frame
			hWnd=GetDlgItem(hwnd, IDC_IMAGE);
			if (NULL != hWnd) {
				hBmp=LoadBitmap(hInstance, MAKEINTRESOURCE(IDB_KBMAP));
				if (NULL != hBmp) {
					SendMessage(hWnd, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hBmp);
				}
			}
			return TRUE;
    } 
    return FALSE; 
} 

BOOL CALLBACK HeatMapProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	// handles a running heatmap display - showing accessed memory
	HWND hWnd=NULL;

    switch (uMsg) 
    { 
        case WM_COMMAND: 
            switch (LOWORD(wParam)) 
            { 
                case IDOK: 
                    // Fall through. 
                 case IDCANCEL: 
                    EndDialog(hwnd, wParam); 
					if (NULL != hHeatBmp) {
						DeleteObject(hHeatBmp);
						hHeatBmp=NULL;
					}
					hHeatMap=NULL;
                    return TRUE; 
            } 
			break;

		case WM_INITDIALOG:
			// create a bitmap for the frame
			hWnd=GetDlgItem(hwnd, IDC_IMAGE);
			if (NULL != hWnd) {
				hHeatBmp=CreateCompatibleBitmap(NULL, 256, 256);
				if (NULL != hHeatBmp) {
					SendMessage(hWnd, STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hHeatBmp);
				}
			}
			return TRUE;
    } 
    return FALSE; 
} 

BOOL CALLBACK BreakPointHelpProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	// nothing fancy here
    switch (uMsg) 
    { 
        case WM_COMMAND: 
            switch (LOWORD(wParam)) 
            { 
                case IDOK: 
                    // Fall through. 
                 case IDCANCEL: 
                    EndDialog(hwnd, wParam); 
					hBrkHlp=NULL;
                    return TRUE; 
            } 
			break;

		case WM_INITDIALOG:
			return TRUE;
    } 
    return FALSE; 
} 

// External references
BOOL CALLBACK TVBoxProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	static double hue=0, sat=0, cont=0, bright=0, sharp=0;
	double thue, tsat, tcont, tbright, tsharp, tmp;

    switch (uMsg) 
    { 
        case WM_COMMAND: 
            switch (LOWORD(wParam)) 
            { 
                case IDOK: 
					TVScanLines=IsDlgButtonChecked(hwnd, IDC_SCANLINES)?1:0;

					// Read the values one more time
					tmp=SendDlgItemMessage(hwnd, IDC_HUE, TBM_GETPOS, 0, 0);
					thue=(tmp-100)/100.0;
					tmp=SendDlgItemMessage(hwnd, IDC_SAT, TBM_GETPOS, 0, 0);
					tsat=(tmp-100)/100.0;
					tmp=SendDlgItemMessage(hwnd, IDC_CONT, TBM_GETPOS, 0, 0);
					tcont=(tmp-100)/100.0;
					tmp=SendDlgItemMessage(hwnd, IDC_BRIGHT, TBM_GETPOS, 0, 0);
					tbright=(tmp-100)/100.0;
					tmp=SendDlgItemMessage(hwnd, IDC_SHARP, TBM_GETPOS, 0, 0);
					tsharp=(tmp-100)/100.0;

					SetTVValues(thue, tsat, tcont, tbright, tsharp);
					InvalidateRect(myWnd, NULL, false);

                    EndDialog(hwnd, wParam); 
					hTVDlg=NULL;
                    return TRUE; 

				case IDCANCEL: 
					// set back to cached values (note: no cancel button on dialog today!)
					SetTVValues(hue, sat, cont, bright, sharp);
                    EndDialog(hwnd, wParam); 
					hTVDlg=NULL;
                    return TRUE; 

				case IDC_RESET:
					SetTVValues(0, 0, 0, 0, 0);
					InvalidateRect(myWnd, NULL, false);
					
					SendDlgItemMessage(hwnd, IDC_HUE, TBM_SETPOS, TRUE, 100);
					SendDlgItemMessage(hwnd, IDC_SAT, TBM_SETPOS, TRUE, 100);
					SendDlgItemMessage(hwnd, IDC_CONT, TBM_SETPOS, TRUE, 100);
					SendDlgItemMessage(hwnd, IDC_BRIGHT, TBM_SETPOS, TRUE, 100);
					SendDlgItemMessage(hwnd, IDC_SHARP, TBM_SETPOS, TRUE, 100);
					return TRUE;

				 case IDC_SCANLINES:
					TVScanLines=IsDlgButtonChecked(hwnd, IDC_SCANLINES)?1:0;
					InvalidateRect(myWnd, NULL, false);
					return TRUE;
            } 
			break;

		case WM_VSCROLL:
			if (NULL != lParam) {
				switch (LOWORD(wParam)) {
					case SB_THUMBPOSITION:
					case SB_THUMBTRACK:
						// High word contains absolute position
						tmp=(HIWORD(wParam)-100)/100.0;
						GetTVValues(&thue, &tsat, &tcont, &tbright, &tsharp);
						if (lParam == (long)GetDlgItem(hwnd, IDC_HUE)) {
							thue=tmp;
						}
						if (lParam == (long)GetDlgItem(hwnd, IDC_SAT)) {
							tsat=tmp;
						}
						if (lParam == (long)GetDlgItem(hwnd, IDC_CONT)) {
							tcont=tmp;
						}
						if (lParam == (long)GetDlgItem(hwnd, IDC_BRIGHT)) {
							tbright=tmp;
						}
						if (lParam == (long)GetDlgItem(hwnd, IDC_SHARP)) {
							tsharp=tmp;
						}
						SetTVValues(thue, tsat, tcont, tbright, tsharp);
						InvalidateRect(myWnd, NULL, false);
						break;

					case SB_BOTTOM:
					case SB_TOP:
					case SB_LINEDOWN:
					case SB_LINEUP:
					case SB_PAGEDOWN:
					case SB_PAGEUP:
					case SB_ENDSCROLL:
						// Need to query for the position (mostly keyboard interface)
						tmp=SendMessage((HWND)lParam, TBM_GETPOS, 0, 0);
						tmp=(tmp-100)/100.0;
						GetTVValues(&thue, &tsat, &tcont, &tbright, &tsharp);
						if (lParam == (long)GetDlgItem(hwnd, IDC_HUE)) {
							thue=tmp;
						}
						if (lParam == (long)GetDlgItem(hwnd, IDC_SAT)) {
							tsat=tmp;
						}
						if (lParam == (long)GetDlgItem(hwnd, IDC_CONT)) {
							tcont=tmp;
						}
						if (lParam == (long)GetDlgItem(hwnd, IDC_BRIGHT)) {
							tbright=tmp;
						}
						if (lParam == (long)GetDlgItem(hwnd, IDC_SHARP)) {
							tsharp=tmp;
						}
						SetTVValues(thue, tsat, tcont, tbright, tsharp);
						InvalidateRect(myWnd, NULL, false);
						break;
				}
			}
			return 0;

		case WM_INITDIALOG:
			// cache the current values in case we cancel
			GetTVValues(&hue, &sat, &cont, &bright, &sharp);

			// load the controls with current values
			SendDlgItemMessage(hwnd, IDC_HUE, TBM_SETRANGE, TRUE, MAKELONG(0,200));
			SendDlgItemMessage(hwnd, IDC_HUE, TBM_SETPOS, TRUE, (int)((hue+1.0)*100));

			SendDlgItemMessage(hwnd, IDC_SAT, TBM_SETRANGE, TRUE, MAKELONG(0,200));
			SendDlgItemMessage(hwnd, IDC_SAT, TBM_SETPOS, TRUE, (int)((sat+1.0)*100));

			SendDlgItemMessage(hwnd, IDC_CONT, TBM_SETRANGE, TRUE, MAKELONG(0,200));
			SendDlgItemMessage(hwnd, IDC_CONT, TBM_SETPOS, TRUE, (int)((cont+1.0)*100));

			SendDlgItemMessage(hwnd, IDC_BRIGHT, TBM_SETRANGE, TRUE, MAKELONG(0,200));
			SendDlgItemMessage(hwnd, IDC_BRIGHT, TBM_SETPOS, TRUE, (int)((bright+1.0)*100));

			SendDlgItemMessage(hwnd, IDC_SHARP, TBM_SETRANGE, TRUE, MAKELONG(0,200));
			SendDlgItemMessage(hwnd, IDC_SHARP, TBM_SETPOS, TRUE, (int)((sharp+1.0)*100));

			SendDlgItemMessage(hwnd, IDC_SCANLINES, BM_SETCHECK, TVScanLines?BST_CHECKED:BST_UNCHECKED, 0);

			return TRUE;

    } 
    return FALSE; 
} 

// read a four digit hex address which might also be a range
bool ReadRange(int *A, int *B, char *str) {
	*A=-1;		// set the failure cause now

	if (*str == '(') {
		// it's a range, so read A and B
		if (2 != sscanf(str, "(%x-%x", A, B)) {
			// invalid, ignore what we got
			*A=-1;
		}
	} else {
		// single value, B to 0
		*B=0;
		if (1 != sscanf(str, "%x", A)) {
			// invalid, ignore what we got
			*A=-1;
		}
	}

	return ((*A)!=-1);
}

// returns a formatted string - not thread safe
const char *FormatBreakpoint(int idx) {
	static char szTmp[32];
	int pos=0;
	
	strcpy(szTmp, "");		// empty it for now

	if (idx < MAX_BREAKPOINTS) {
		// write the prefix, if any
		switch (BreakPoints[idx].Type) {
			case BREAK_ACCESS:
				szTmp[pos++]='*';
				break;

			case BREAK_WRITE:
				szTmp[pos++]='>';
				break;

			case BREAK_READ:
				szTmp[pos++]='<';
				break;

			case BREAK_EQUALS_WORD:
				szTmp[pos++]='W';
				break;

			case BREAK_EQUALS_BYTE:
				szTmp[pos++]='M';
				break;

			case BREAK_EQUALS_VDP:
				szTmp[pos++]='V';
				break;

			case BREAK_EQUALS_VDPREG:
				szTmp[pos++]='U';
				break;

			case BREAK_EQUALS_REGISTER:
				szTmp[pos++]='R';
				break;

			case BREAK_RUN_TIMER:
				szTmp[pos++]='T';
				break;
		}

		// write the address or range
		switch (BreakPoints[idx].Type) {
			case BREAK_EQUALS_VDPREG:
			case BREAK_EQUALS_REGISTER:
				pos+=sprintf(&szTmp[pos], "%d", BreakPoints[idx].A);
				break;

			default:
				if (BreakPoints[idx].B != 0) {
					pos+=sprintf(&szTmp[pos], "(%04X-%04X)", BreakPoints[idx].A, BreakPoints[idx].B);
				} else {
					pos+=sprintf(&szTmp[pos], "%04X", BreakPoints[idx].A);
				}
		}

		// write the data, if needed
		switch (BreakPoints[idx].Type) {
			case BREAK_EQUALS_WORD:
			case BREAK_EQUALS_REGISTER:
				// a word
				pos+=sprintf(&szTmp[pos], "=%04X", BreakPoints[idx].Data);
				break;

			case BREAK_EQUALS_BYTE:
			case BREAK_EQUALS_VDP:
			case BREAK_EQUALS_VDPREG:
				// a byte
				pos+=sprintf(&szTmp[pos], "=%02X", BreakPoints[idx].Data);
				break;
		}
	}

	return szTmp;
}

// true if parsed successfully
// Note that buf1 is modified in place!
bool AddBreakpoint(char *buf1) {
	bool bRet=false;
	int A,B;

	// first check for a PC - no prefix
	if (ReadRange(&A, &B, buf1)) {
		A&=0xffff;
		B&=0xffff;
		BreakPoints[nBreakPoints].Type=BREAK_PC;
		BreakPoints[nBreakPoints].A=A;
		BreakPoints[nBreakPoints].B=B;
		strcpy(buf1, FormatBreakpoint(nBreakPoints));
		nBreakPoints++;
		bRet=true;
	} else {
		int nType=BREAK_NONE;
		int nTmp=0;
		int nData=0;
		char *pTmp;
		
		buf1[0]=toupper(buf1[0]);
		switch (buf1[0]) {
		case '*':	// memory address access (read or write)
			nType=BREAK_ACCESS;
			break;
		case '>':	// write to address
			nType=BREAK_WRITE;
			break;
		case '<':	// read from address
			nType=BREAK_READ;
			break;
		case 'W':	// Word =
			nType=BREAK_EQUALS_WORD;
			pTmp=strchr(buf1, '=');
			if (NULL == pTmp) {
				nType=BREAK_NONE;
			} else {
				if (1 != sscanf(pTmp+1, "%x", &nData)) {
					nType=BREAK_NONE;
				}
			}
			break;
		case 'M':	// Memory = 
			nType=BREAK_EQUALS_BYTE;
			pTmp=strchr(buf1, '=');
			if (NULL == pTmp) {
				nType=BREAK_NONE;
			} else {
				if (1 != sscanf(pTmp+1, "%x", &nData)) {
					nType=BREAK_NONE;
				}
			}
			break;
		case 'U':	// VDP register =
			nType=BREAK_EQUALS_VDPREG;
			pTmp=strchr(buf1, '=');
			if (NULL == pTmp) {
				nType=BREAK_NONE;
			} else {
				if (1 != sscanf(pTmp+1, "%x", &nData)) {
					nType=BREAK_NONE;
				}
			}
			break;
		case 'V':	// VDP memory =
			nType=BREAK_EQUALS_VDP;
			pTmp=strchr(buf1, '=');
			if (NULL == pTmp) {
				nType=BREAK_NONE;
			} else {
				if (1 != sscanf(pTmp+1, "%x", &nData)) {
					nType=BREAK_NONE;
				}
			}
			break;
		case 'R':	// Register =
			nType=BREAK_EQUALS_REGISTER;
			pTmp=strchr(buf1, '=');
			if (NULL == pTmp) {
				nType=BREAK_NONE;
			} else {
				// register is the only one that's read as decimal!
				if (1 != sscanf(pTmp+1, "%d", &nData)) {
					nType=BREAK_NONE;
				}
			}
			break;

		case 'T':	// Timing run
			nType=BREAK_RUN_TIMER;
			break;
		}

		if (nType != BREAK_NONE) {
			if (ReadRange(&A, &B, &buf1[1])) {
				A&=0xffff;
				B&=0xffff;
				// eliminate a few special cases
				switch (nType) {
					case BREAK_RUN_TIMER:
						if (B==0) {
							nType = BREAK_NONE;		// invalid without range
						}
						break;

					case BREAK_EQUALS_REGISTER:
					case BREAK_EQUALS_VDPREG:
						if (B!=0) {
							nType = BREAK_NONE;		// range not legal on registers
						}
						break;
				}

				if (nType != BREAK_NONE) {
					// still valid, now we need to save and format it
					BreakPoints[nBreakPoints].Type=nType;
					BreakPoints[nBreakPoints].A=A;
					BreakPoints[nBreakPoints].B=B;
					BreakPoints[nBreakPoints].Data=nData;
					strcpy(buf1, FormatBreakpoint(nBreakPoints));
					nBreakPoints++;
					bRet=true;
				}
			}
		}
	}

	return bRet;
}

void LaunchDebugWindow() {
	DoPause();
	Sleep(100);
	if (NULL == dbgWnd) {
		dbgWnd=CreateDialog(NULL, MAKEINTRESOURCE(IDD_DEBUG), myWnd, DebugBoxProc);
	}
	ShowWindow(dbgWnd, SW_SHOW);
	Sleep(100);
	DoPlay();
}

BOOL CALLBACK DebugBoxProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	char buf1[80];
	int idx;

    switch (uMsg) 
    { 
        case WM_COMMAND: 
            switch (LOWORD(wParam)) 
            { 
				case IDOK: 
				case IDCANCEL: 
				case IDC_CLOSE:
                    EndDialog(hwnd, wParam); 
					dbgWnd=NULL;
                    return TRUE; 

				case IDC_RADIO1:
					bDebugDirty=true;
					// fall through
				case IDC_RADIO2:
				case IDC_RADIO3:
				case IDC_RADIO4:
				case IDC_RADIO5:
					nMemType=LOWORD(wParam)-IDC_RADIO1;
					SetDlgItemText(hwnd, IDC_ADDRESS, szTopMemory[nMemType]);
					SetEvent(hDebugWindowUpdateEvent);
					break;

				case IDC_ADDBREAK:
					{
						// add a new breakpoint
						if (nBreakPoints >= MAX_BREAKPOINTS) {
							Beep(550,100);
							break;
						}
						if (SendDlgItemMessage(hwnd, IDC_COMBO1, WM_GETTEXT, 80, (LPARAM)buf1) > 0) {
							if (AddBreakpoint(buf1)) {
								SendDlgItemMessage(hwnd, IDC_COMBO1, CB_ADDSTRING, 0, (LPARAM)buf1);
							}
						}
					}
					break;

				case IDC_REMOVEBREAK:
					{
						int n;
						while (CB_ERR != (n=SendDlgItemMessage(hwnd, IDC_COMBO1, CB_GETCURSEL, 0, 0))) {
							SendDlgItemMessage(hwnd, IDC_COMBO1, CB_DELETESTRING, n, 0);
						}
						// rebuild the breakpoint list
						nBreakPoints=0;
						n=SendDlgItemMessage(hwnd, IDC_COMBO1, CB_GETCOUNT, 0, 0);
						while (n > 0) {
							char buf[32];
							strcpy(buf, "");
							SendDlgItemMessage(hwnd, IDC_COMBO1, CB_GETLBTEXT, n-1, (LPARAM)buf);
							if (strlen(buf)) {
								AddBreakpoint(buf);
							}
							n--;
						}
					}
					break;

				case IDC_BREAKHELP:
					{
						if (NULL == hBrkHlp) {
							// create a modeless dialog to show the breakpoint help
							hBrkHlp=CreateDialog(NULL, MAKEINTRESOURCE(IDD_BRKHLP), hwnd, BreakPointHelpProc);
							ShowWindow(hBrkHlp, SW_SHOW);
						}
					}
					break;

// max lines is 34, each line is 8 bytes
#define LINES_TO_STEP 32
				case IDC_NEXT:
					switch (nMemType) {
						case 2:		// CPU
						case 3:		// VDP
						case 4:		// GROM
							// step forward in memory 
							{
								int x;
								// validate it
								if (1 == sscanf(szTopMemory[nMemType], "%X", &x)) {
									// it is a number, not a register
									x+=(LINES_TO_STEP*8);
									sprintf(szTopMemory[nMemType], "%04X", x);
									SetDlgItemText(hwnd, IDC_ADDRESS, szTopMemory[nMemType]);
								}
								// refresh the dialog
								SetEvent(hDebugWindowUpdateEvent);
							}
							break;
					}
					break;

				case IDC_PREV:
					switch (nMemType) {
						case 2:		// CPU
						case 3:		// VDP
						case 4:		// GROM
							// step back in memory 
							{
								int x;
								// validate it
								if (1 == sscanf(szTopMemory[nMemType], "%X", &x)) {
									// it is a number, not a register
									x-=(LINES_TO_STEP*8);
									sprintf(szTopMemory[nMemType], "%04X", x);
									SetDlgItemText(hwnd, IDC_ADDRESS, szTopMemory[nMemType]);
								}
								// refresh the dialog
								SetEvent(hDebugWindowUpdateEvent);
							}
							break;
					}
					break;

				case IDC_APPLY:
					switch (nMemType) {
						case 1:		// disassembly
							{
								char buf[128];
								int x;
								memset(buf, 0, sizeof(buf));
								GetDlgItemText(hwnd, IDC_ADDRESS, buf, 128);
								_strupr(buf);
								if (1 == sscanf(buf, "PC=%X", &x)) {
									// make a change to the Program Counter
									PC=x;
									// refresh the dialog
									SetEvent(hDebugWindowUpdateEvent);
								}
								if (1 == sscanf(buf, "WP=%X", &x)) {
									// make a change to the Workspace Pointer
									WP=x;
									// refresh the dialog
									SetEvent(hDebugWindowUpdateEvent);
								}
							}
							break;

						case 2:		// CPU
						case 3:		// VDP
						case 4:		// GROM
							{
								char buf[128];
								int x,y;
								int ok,byte,romok;
								romok=0;
								memset(buf, 0, sizeof(buf));
								GetDlgItemText(hwnd, IDC_ADDRESS, buf, 128);
								_strupr(buf);								

								// hacky little duplication of WP/PC tests
								{
									if (1 == sscanf(buf, "PC=%X", &x)) {
										// make a change to the Program Counter
										PC=x;
										// refresh the dialog
										SetEvent(hDebugWindowUpdateEvent);
										break;
									}
									if (1 == sscanf(buf, "WP=%X", &x)) {
										// make a change to the Workspace Pointer
										WP=x;
										// refresh the dialog
										SetEvent(hDebugWindowUpdateEvent);
										break;
									}
								}

								// validate it
								if (strchr(buf, '=')) {
									// this is a write command!
									// remove leading whitespace that might screw up the register check
									while (isspace(buf[0])) {
										memmove(&buf[0], &buf[1], sizeof(buf)-1);
									}
									if ((2 == sscanf(buf, "%X=%X", &x, &y)) || (2 == sscanf(buf, "R%d=%X", &x, &y))) {
										// make a change -- convert x if register
										if (buf[0]=='R') {
											if ((nMemType == 2) || (nMemType == 3)) {  // CPU or VDP
												x=-(x+1);	// registers are -1 (R0) to -16 (R15) (to -8 (R7) for VDP)
											} else {
												// no registers elsewhere
												MessageBox(hwnd, "No registers on this memory type", "Classic99 Debugger", MB_ICONSTOP);
												break;
											}
										}
										// check if y is byte or word
										char *p=strchr(buf,'=');
										p++;
										byte=0;
										while (isspace(*p)) p++;
										while (isalnum(*p)) {
											byte++;
											p++;
										}
										if ((byte > 2) && (nMemType == 2)) {
											byte = 0;		// no, word -- only the CPU supports this
											y&=0xffff;
										} else {
											byte = 1;		// yes, byte
											y&=0xff;
										}
										// do various warnings/errors
										ok=IDYES;
										switch (nMemType) {
											case 2:		// CPU
												if (x < -16) {
													MessageBox(hwnd, "Out of range for CPU registers", "Classic99 Debugger", MB_ICONSTOP);
													ok=IDNO;
												} else if (ROMMAP[x]) {
													ok=MessageBox(hwnd, "Modify ROM? (Bank 0 only, non-permanent)\nif you answer no, the write will be performed normally.", "Classic99 Debugger", MB_YESNO|MB_ICONQUESTION|MB_DEFBUTTON2);
													if (ok == IDYES) {
														romok=1;
													} else {
														romok=0;
													}
													// allow it either way
													ok=IDYES;
												} else {
													ok=IDYES;
												}
												break;
											case 3:		// VDP
												if (x > 0x3FFF) {
													MessageBox(hwnd, "Out of range for VDP address", "Classic99 Debugger", MB_ICONSTOP);
													ok=IDNO;
												}
												if (x < -8) {
													MessageBox(hwnd, "Out of range for VDP registers", "Classic99 Debugger", MB_ICONSTOP);
													ok=IDNO;
												}
												break;
											case 4:		// GROM
												ok=MessageBox(hwnd, "Modify GROM? (Base 0 only, non-permanent)", "Classic99 Debugger", MB_YESNO|MB_ICONQUESTION|MB_DEFBUTTON2);
												break;
											default:
												ok=IDNO;
												break;
										}
										if (ok == IDYES) {
											// CPU writes /are/ allowed to cause side effects!
											switch (nMemType) {
												case 2:		// CPU again
													if (x < 0) {
														// convert negative register index to address
														x=WP - ((x+1)*2);
													}
													if (byte) {
														if ((ROMMAP[x]) && (romok)) {
															// need to be direct
															staticCPU[x]=y;
														} else {
															// do this so side effects work
															wcpubyte(x, y);
														}
													} else {
														if ((ROMMAP[x]) && (romok)) {
															// need to be direct
															staticCPU[x]=y>>8;
															staticCPU[x+1]=y&0xff;
														} else {
															// do this so side effects work
															wrword(x, y);
														}
													}
													break;

												case 3:		// VDP again
													if (x < 0) {
														// it's a register
														VDPREG[(-x)-1] = y;
													} else {
														VDP[x] = y;
													}
													// force VDP to update
													SetEvent(DisplayEvent);
													break;

												case 4:		// GROM 
													// only base zero handled
													GROMBase[0].GROM[x]=y;
													break;
											}
										}
									} else {
										MessageBox(hwnd, "Use XXXX=XX to set a value (both values in hex)\nor Rx=XX for a register (register index in decimal)\nWP=xxxx and PC=xxxx are also legal.", "Unrecognized string", MB_ICONASTERISK);
									}
								} else {
									if ((1 == sscanf(buf, "%X", &x)) || (tolower(buf[0])=='r')) {
										strcpy(szTopMemory[nMemType], buf);
									}
								}
								// refresh the dialog
								SetEvent(hDebugWindowUpdateEvent);
							}
							break;
					}
					break;

				case IDC_EDITMEM:
				;

				case IDC_PAUSE:
					if (0 == max_cpf) {
						// already paused, restore
						DoPlay();
					} else {
						// running normal, so pause
						DoPause();
					}
					break;

				case IDC_STEP:
					DoStep();
					break;

				case IDC_STEP_OVER:
					DoStepOver();
					break;

				case IDC_PLAY:
					DoPlay();
					break;

				case IDC_FASTFORWARD:
					DoFastForward();
					break;

				case IDC_VDPCHARS:
					VDPDebug=!VDPDebug;
					if (DisplayEvent) {
						SetEvent(DisplayEvent);
					}
					break;

				case IDC_LOADINT:
					DoLoadInterrupt();
					break;

				case IDC_DUMPMEMORY:
					DoMemoryDump();
					break;

				case IDC_FREEZE:
					if (bFrozenText) {
						bFrozenText=false;
						SendDlgItemMessage(hwnd, IDC_FREEZE, WM_SETTEXT, NULL, (LPARAM)"Freeze");
					} else {
						bFrozenText=true;
						SendDlgItemMessage(hwnd, IDC_FREEZE, WM_SETTEXT, NULL, (LPARAM)"UnFreeze");
					}
					break;

				case IDC_REDRAW:
					bDebugDirty=true;
					SetEvent(hDebugWindowUpdateEvent);
					break;

				case IDC_CLEAR:
					if (0 == nMemType) {
						// clear debug buffer
						memset(lines, 0, sizeof(lines));
						bDebugDirty=true;
						SetEvent(hDebugWindowUpdateEvent);
					}
					break;
            } 
			return TRUE;

		case WM_INITDIALOG:
			CheckRadioButton(hwnd, IDC_RADIO1, IDC_RADIO5, IDC_RADIO1);
			nMemType=0;
			SendDlgItemMessage(hwnd, IDC_MAINEDIT, WM_SETTEXT, NULL, (LPARAM)"");

			// breakpoints
			SendDlgItemMessage(hwnd, IDC_COMBO1, CB_RESETCONTENT, NULL, NULL);
			for (idx=0; idx<nBreakPoints; idx++) {
				char buf[128];
				strcpy(buf, FormatBreakpoint(idx));
				SendDlgItemMessage(hwnd, IDC_COMBO1, CB_ADDSTRING, NULL, (LPARAM)buf);
			}
			SendDlgItemMessage(hwnd, IDC_COMBO1, CB_LIMITTEXT, 16, NULL);
			// fall through
		case WM_APP:
			// refresh the dialog
			SetEvent(hDebugWindowUpdateEvent);
			return TRUE;
    } 
    return FALSE; 
} 

void DebugUpdateThread(void*) {
	static int nOldPC=-1;
	static int nOldMemType=-1;
	static char szOldMemory[5][32] = { "", "", "", "", "" };
	static CString csOut;

	char buf1[80], buf2[80];
	int myPC, idx, idx2;
	int tmpPC;
	HWND hWnd;

	while (!quitflag) {
		if (WAIT_TIMEOUT == WaitForSingleObject(hDebugWindowUpdateEvent, 500)) {
			// timeout every so often to recheck quitflag
			continue;
		}

		if (NULL == dbgWnd) {
			continue;
		}

		// Helps a bit -- not much can change unless the PC does ;)
		if ((nOldPC == PC) && (nOldMemType == nMemType) && (0 == memcmp(szOldMemory, szTopMemory, sizeof(szOldMemory)))) {
			continue;
		}

		nOldPC=PC;
		nOldMemType=nMemType;

		// Debug window
		hWnd=GetDlgItem(dbgWnd, IDC_MAINEDIT);
		if (NULL != hWnd) {
			csOut="";

			if (!bFrozenText) {
				switch (nMemType) {
					case 0:		// debug
						if (bDebugDirty) {
							for (idx=1; idx<34; idx++) {
								csOut+=lines[idx];
								csOut+="\r\n";
							}
							bDebugDirty=false;
						}
						break;

					case 1:		// disassembly
						{
							for (idx=0; idx<20; idx++) {
								Dasm9900(buf2, disasm[idx]);
								sprintf(buf1, "  %04X   %-30s\r\n", disasm[idx], buf2);
								csOut+=buf1;
							}

							myPC=PC;
							
							tmpPC = myPC;
							myPC += Dasm9900(buf2, myPC);
							myPC&=0xffff;

							sprintf(buf1, "> %04X   %-30s\r\n", tmpPC, buf2);
							csOut+=buf1;

							for (idx=0; idx<13; idx++) {
								tmpPC = myPC;
								myPC += Dasm9900(buf2, myPC);
								myPC&=0xffff;
								sprintf(buf1, "  %04X   %-30s\r\n", tmpPC, buf2);
								csOut+=buf1;
							}

						}
						break;

					case 2:		// CPU Memory
						{
							// CPU must not call the read function, as it may call the memory
							// mapped devices and affect them. This code does not take into
							// account scratchpad or cartridge ROM banks
							char buf3[32];
							int c;
							strncpy(buf3, szTopMemory[nMemType], 32);
							buf3[31]='\0';
							if (1 != sscanf(buf3, "%X", &tmpPC)) {
								tmpPC=0;
								if (tolower(buf3[0])=='r') {
									c=atoi(&buf3[1])*2;
									tmpPC=GetSafeByte(WP+c)*256 + GetSafeByte(WP+c+1);
								}
							}
							tmpPC&=0xffff;
							for (idx2=0; idx2<34; idx2++) {
								sprintf(buf1, "%04X: ", tmpPC);
								strcpy(buf3, "");
								for (idx=0; idx<8; idx++) {
									c=GetSafeByte(tmpPC++);
									tmpPC&=0xffff;
									sprintf(buf2, "%02X ", c);
									strcat(buf1, buf2);
									if ((c>=32)&&(c<=126)) {
										buf2[0]=c;
									} else {
										buf2[0]='.';
									}
									buf2[1]='\0';
									strcat(buf3, buf2);
								}
								strcat(buf1, buf3);
								csOut+=buf1;
								csOut+="\r\n";
							}
						}
						break;

					case 3:		// VDP Memory
						{
							// VDP and GROM *must not* call the read functions, as it will
							// change the address!! (But, since we don't do bank switching in them...)
							char buf3[32];
							int c;
							strncpy(buf3, szTopMemory[nMemType], 32);
							buf3[31]='\0';
							if (1 != sscanf(buf3, "%X", &tmpPC)) {
								tmpPC=0;
								if (tolower(buf3[0])=='r') {
									c=atoi(&buf3[1])*2;
									tmpPC=GetSafeByte(WP+c)*256 + GetSafeByte(WP+c+1);
								}
							}
							tmpPC&=0x3fff;
							for (idx2=0; idx2<34; idx2++) {
								sprintf(buf1, "%04X: ", tmpPC);
								strcpy(buf3, "");
								for (idx=0; idx<8; idx++) {
									c=VDP[tmpPC++];
									tmpPC&=0x3fff;
									sprintf(buf2, "%02X ", c);
									strcat(buf1, buf2);
									if ((c>=32)&&(c<=126)) {
										buf2[0]=c;
									} else {
										buf2[0]='.';
									}
									buf2[1]='\0';
									strcat(buf3, buf2);
								}
								strcat(buf1, buf3);
								csOut+=buf1;
								csOut+="\r\n";
							}
						}
						break;

					case 4:		// GROM
							// VDP and GROM *must not* call the read functions, as it will
							// change the address!! (But, since we don't do bank switching in them...)
							// Bug: GROM base 0 only
							char buf3[32];
							int c;
							strncpy(buf3, szTopMemory[nMemType], 32);
							buf3[31]='\0';
							if (1 != sscanf(buf3, "%X", &tmpPC)) {
								tmpPC=0;
								if (tolower(buf3[0])=='r') {
									c=atoi(&buf3[1])*2;
									tmpPC=GetSafeByte(WP+c)*256 + GetSafeByte(WP+c+1);
								}
							}
							for (idx2=0; idx2<34; idx2++) {
								sprintf(buf1, "%04X: ", tmpPC);
								strcpy(buf3, "");
								for (idx=0; idx<8; idx++) {
									c=GROMBase[0].GROM[tmpPC++];
									tmpPC&=0xffff;
									sprintf(buf2, "%02X ", c);
									strcat(buf1, buf2);
									if ((c>=32)&&(c<=126)) {
										buf2[0]=c;
									} else {
										buf2[0]='.';
									}
									buf2[1]='\0';
									strcat(buf3, buf2);
								}
								strcat(buf1, buf3);
								csOut+=buf1;
								csOut+="\r\n";
							}
							break;
				}
				if (!csOut.IsEmpty()) {
					SendMessage(hWnd, WM_SETTEXT, NULL, (LPARAM)(LPCSTR)csOut);
				}
			}
		}

		hWnd=GetDlgItem(dbgWnd, IDC_SECONDEDIT);
		if (NULL != hWnd) {
			int val;

			csOut="";

			if (!bFrozenText) {
				// prints the register information in a single edit control
				// spacing: <5 label><1 space><4 value><4 spaces><5 label><1 space><4 value>
				for (idx=0; idx<8; idx++) {
					int val=GetSafeWord(WP+idx*2);
					int val2=GetSafeWord(WP+(idx+8)*2);
					sprintf(buf1, " R%2d  %04X     R%2d  %04X\r\n", idx, val, idx+8, val2);
					csOut+=buf1;
				}

				csOut+="\r\n";

				// VDP registers and associated registers beside them
				for (idx=0; idx<8; idx++) {
					val=VDPREG[idx];
					sprintf(buf1, "VDP%d  %02X      ", idx, val);
					csOut+=buf1;
					switch (idx) {
						case 0:	sprintf(buf1, " VDP  %04X\r\n", VDPADD); break;
						case 1: sprintf(buf1, " GROM %04X\r\n", GROMBase[0].GRMADD); break;
						case 2: sprintf(buf1, "VDPST %02X\r\n", VDPS); break;
						case 3: sprintf(buf1, "  PC  %04X\r\n", PC); break;
						case 4: sprintf(buf1, "  WP  %04X\r\n", WP); break;
						case 5: sprintf(buf1, "  ST  %04X\r\n", ST); break;
						case 6: sprintf(buf1, " Bank %04X\r\n", (xb<<8)|(xbBank)); break;
						case 7: sprintf(buf1, " DSR  %04X\r\n", nCurrentDSR&0xffff); break;
						default: strcpy(buf1, "\r\n");
					}
					csOut+=buf1;
				}

				csOut+="\r\n";

				// VDP tables
				sprintf(buf1, " SIT  %04X\r\n", SIT);
				csOut+=buf1;
				sprintf(buf1, " SDT  %04X     SAL  %04X\r\n", SDT, SAL);
				csOut+=buf1;
				sprintf(buf1, " PDT  %04X     Size %04X\r\n", PDT, PDTsize);
				csOut+=buf1;
				sprintf(buf1, "  CT  %04X     Size %04X\r\n", CT, CTsize);
				csOut+=buf1;

				csOut+="\r\n";

				// break down the status registers
				val=VDPS;
				sprintf(buf1, "VDPST: %s %s %s", (val&VDPS_INT)?"INT":"   ", (val&VDPS_5SPR)?"5SP":"   ", (val&VDPS_SCOL)?"COL":"   ");
				csOut+=buf1;
				if (val&VDPS_5SPR) {
					sprintf(buf1, "5thSP: %X\r\n", VDPS&(~(VDPS_INT|VDPS_5SPR|VDPS_SCOL)));
				} else {
					strcpy(buf1, "\r\n");
				}
				csOut+=buf1;

				csOut+="\r\n";

				val=ST;
				sprintf(buf1, "  ST : %s %s %s %s %s %s %s\r\n", (val&BIT_LGT)?"LGT":"   ", (val&BIT_AGT)?"AGT":"   ", (val&BIT_EQ)?"EQ":"  ",
					(val&BIT_C)?"C":" ", (val&BIT_OV)?"OV":"  ", (val&BIT_OP)?"OP":"  ", (val&BIT_X)?"X":" ");
				csOut+=buf1;
				sprintf(buf1, " MASK: %X\r\n", val&INTMASK);
				csOut+=buf1;

				csOut+="\r\n";

				// Sound chip
				sprintf(buf1, " 9919 %03X %03X %03X %X\r\n", nRegister[0], nRegister[1], nRegister[2], nRegister[3]);
				csOut+=buf1;

				sprintf(buf1, " VOL   %X   %X   %X  %X\r\n", nVolume[0], nVolume[1], nVolume[2], nVolume[3]);
				csOut+=buf1;

				SendMessage(hWnd, WM_SETTEXT, NULL, (LPARAM)(LPCSTR)csOut);
			}
		}
	}
}
